@@ -107,6 +107,17 @@ sub KEY_create_rsa {
     return $key;
 }
 
+if (defined &Net::SSLeay::EC_KEY_generate_key) {
+    push @EXPORT,'KEY_create_ec';
+    *KEY_create_ec = sub {
+	my $curve = shift || 'prime256v1';
+	my $key = Net::SSLeay::EVP_PKEY_new();
+	my $ec = Net::SSLeay::EC_KEY_generate_key($curve);
+	Net::SSLeay::EVP_PKEY_assign_EC_KEY($key,$ec);
+	return $key;
+    }
+}
+
 # extract information from cert
 my %gen2i = qw( OTHERNAME 0 EMAIL 1 DNS 2 X400 3 DIRNAME 4 EDIPARTY 5 URI 6 IP 7 RID 8 );
 my %i2gen = reverse %gen2i;
@@ -370,10 +381,15 @@ sub CERT_create {
 	=> join(",",sort keys %ext_key_usage) if %ext_key_usage;
     Net::SSLeay::P_X509_add_extensions($cert, $issuer_cert, @ext);
 
+    my %have_ext;
+    for(my $i=0;$i<@ext;$i+=2) {
+	$have_ext{ $ext[$i] }++
+    }
     for my $ext (@{ $args{ext} || [] }) {
 	my $nid = $ext->{nid}
 	    || $ext->{sn} && Net::SSLeay::OBJ_sn2nid($ext->{sn})
 	    || croak "cannot determine NID of extension";
+	$have_ext{$nid} and next;
 	my $val = $ext->{data};
 	if ($nid == 177) {
 	    # authorityInfoAccess:
@@ -506,6 +522,10 @@ Each loaded or created cert and key must be freed to not leak memory.
 
 Creates an RSA key pair, bits defaults to 2048.
 
+=item * KEY_create_ec(curve) -> key
+
+Creates an EC key, curve defaults to C<prime256v1>.
+
 =item * CERT_asHash(cert,[digest_algo]) -> hash
 
 Extracts the information from the certificate into a hash and uses the given
